# -*- coding: utf-8 -*-
"""
module for generating input parameters
"""

import random
import string
import sys

import ast

# could potentially be unicode instead
CHARS = string.printable


def _str_gen(length=10):
    while True:
        yield ''.join(random.choice(CHARS) for _ in xrange(length))


def _long_gen():
    """
    long generator.
    because it isn't bounded, will be ever increasing with random gap
    """
    yield 0
    num = 1
    while True:
        yield num
        yield -num
        num += random.randint(0, num)


def _int_gen(lower=-sys.maxint - 1, upper=sys.maxint):
    if lower is None:
        lower = -sys.maxint - 1
    if upper is None:
        upper = sys.maxint
    assert type(lower) == int, "lower bound must be type int, got {}".format(lower)
    assert type(upper) == int, "upper bound must be type int, got {}".format(upper)
    while True:
        # note that this is inclusive
        yield random.randint(lower, upper)


def _float_gen():
    while True:
        yield random.random()


SUPPORTED_TYPES = (int, str, float)

# compare ops
COMPARE_OPS = {ast.Eq, ast.NotEq,
               ast.Lt, ast.GtE,
               ast.Gt, ast.LtE,
               ast.Is, ast.IsNot,
               ast.In, ast.NotIn}

SWAPPED = {ast.Eq: ast.Eq,
           ast.Lt: ast.Gt,
           ast.LtE: ast.GtE}

SWAPPED.update({v: k for k, v in SWAPPED.copy().iteritems()})

REVERSE = {ast.Eq: ast.NotEq,
           ast.Lt: ast.GtE,
           ast.Gt: ast.LtE,
           ast.Is: ast.IsNot,
           ast.In: ast.NotIn}

REVERSE.update({v: k for k, v in REVERSE.copy().iteritems()})


class Visitor(ast.NodeVisitor):
    """
    a node visitor that keeps track of only comparators

    TODO: extend and support other operators
    """

    def __init__(self):
        self.statements = {}

    def _process_compare(self, left, op, right):
        op = op.__class__
        if left.__class__ is ast.Num and right.__class__ is ast.Num:
            return
        if left.__class__ is ast.Name and right.__class__ is ast.Name:
            key = (left.id, right.id)
            val = self.statements.get(key, [])
            val.append(op)
            self.statements[key] = val
            return
        if left.__class__ is ast.Num:
            left, right = right, left
            op = SWAPPED[op]
        key = left.id
        val = self.statements.get(key, [])
        val.append((op, right.n))
        self.statements[key] = val


    def visit_Compare(self, node):
        """
        visit all compare expressions
        """
        assert all(op.__class__ in COMPARE_OPS for op in node.ops)
        left = node.left
        for op, right in zip(node.ops, node.comparators):
            self._process_compare(left, op, right)
            left = right


def _parse_lambda_expr(args, expr):
    visitor = Visitor()
    visitor.visit(expr)
    return visitor.statements


def _parse_lambda(lmd):
    arguments, body = lmd.args, lmd.body
    args = arguments.args
    assert arguments.vararg is None
    assert arguments.kwarg is None
    assert arguments.defaults == []
    return _parse_lambda_expr(args, body)


def _parse_condition(cond_str):
    node = ast.parse(cond_str)
    child, = list(ast.iter_child_nodes(node))
    return _parse_lambda(child.value)


def _type_to_generator(arg_type, lower=None, upper=None):
    if arg_type == int:
        return _int_gen(lower=lower, upper=upper)
    elif arg_type == long:
        return _long_gen()


def _decide_range(condition_list):
    lower = None
    upper = None
    for condition in condition_list:
        op, num = condition
        if op == ast.LtE:
            if upper is None or upper > num:
                upper = num
        elif op == ast.GtE:
            if lower is None or lower < num:
                lower = num
        else:
            assert False, op
    return lower, upper


def make_generators(args, conditions):
    ret = []
    for i in xrange(len(args)):
        name = args[i].name
        arg_type = args[i].type
        if name in conditions:  # with some constraints
            condition_list = conditions[name]
            lower, upper = _decide_range(condition_list)
            print "param: {}, lower: {}, upper: {}".format(name, lower, upper)
            ret.append(_type_to_generator(arg_type, lower=lower, upper=upper))
        else:  # empty, all range
            ret.append(_type_to_generator(arg_type))
    return ret


class DataGenerator:
    """
    the data generator will return an unbounded iterator of argument lists
    """

    def __init__(self, method_prop):
        prop = method_prop.prop
        self.pre = prop.pre
        conditions = _parse_condition(prop.ast['pre'])
        self.generators = make_generators(method_prop.args, conditions)

    def __iter__(self):
        while True:
            item = []
            for gen in self.generators:
                item.append(gen.next())
            yield tuple(item)
